# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from datetime import datetime
from typing import Optional, TypeVar

TransferType = TypeVar("TransferType", bound="Transfer")


@dataclass
class Transfer:
    id: str
    issuer_address: str
    issuer_identity_index: Optional[int]
    issuer_identity_name: Optional[str]
    receiver_address: str
    receiver_identity_index: Optional[int]
    receiver_identity_name: Optional[str]
    amount: int
    timestamp: datetime
    comment: Optional[str]
    comment_type: Optional[str]

    def __eq__(self, other):
        """
        Test equality on address

        :param other: Account instance
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash(f"{self.id}")



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from tikka.domains.entities.technical_committee import (
    TechnicalCommitteeMember,
    TechnicalCommitteeProposal,
)
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.network.network import NetworkInterface
from tikka.interfaces.adapters.repository.repository import RepositoryInterface
from tikka.libs.keypair import Keypair


class TechnicalCommittee:
    """
    TechnicalCommittee domain class
    """

    def __init__(
        self,
        repository: RepositoryInterface,
        network: NetworkInterface,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init TechnicalCommittee domain

        :param repository: RepositoryInterface instance
        :param network: NodeTechnicalCommitteeInterface instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.repository = repository
        self.network = network
        self.event_dispatcher = event_dispatcher

    def network_update_members(self) -> None:
        """
        Return all members of Technical Committee from network

        :return:
        """
        members = self.network.node.technical_committee.members()
        if self.network.indexer.connection.is_connected():
            indice = [
                member.identity_index
                for member in members
                if member.identity_index is not None
            ]
            identity_names = self.network.indexer.identities.get_identity_names(indice)
            for member in members:
                if member.identity_index in identity_names:
                    member.identity_name = identity_names[member.identity_index]

        self.repository.technical_committee_members.set_list(members)

    def network_update_proposals(self) -> None:
        """
        Return all proposals with voting infos of Technical Committee from network

        :return:
        """
        proposals = self.network.node.technical_committee.proposals()
        self.repository.technical_committee_proposals.set_list(proposals)

    def network_vote(
        self, keypair: Keypair, proposal: TechnicalCommitteeProposal, vote: bool
    ) -> None:
        """
        Send Technical Committee Vote for proposal from Keypair

        :param keypair: Keypair instance
        :param proposal: TechnicalCommitteeProposal instance
        :param vote: True or False
        :return:
        """
        return self.network.node.technical_committee.vote(keypair, proposal, vote)

    def list_members(self):
        """
        Return list of technical committee members

        :return:
        """
        return self.repository.technical_committee_members.list()

    def set_members_list(self, members: List[TechnicalCommitteeMember]) -> None:
        """
        Set list of technical committee members with members

        :return:
        """
        self.repository.technical_committee_members.set_list(members)

    def list_proposals(self):
        """
        Return list of technical committee proposals

        :return:
        """
        return self.repository.technical_committee_proposals.list()

    def set_proposals_list(self, proposals: List[TechnicalCommitteeProposal]) -> None:
        """
        Set list of technical committee proposals with proposals

        :return:
        """
        self.repository.technical_committee_proposals.set_list(proposals)

    def list_member_addresses(self):
        """
        Return list of technical committee member addresses

        :return:
        """
        return [
            technical_committee_member.address
            for technical_committee_member in self.repository.technical_committee_members.list()
        ]



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.resources.gui.windows.category_add_rc import Ui_addCategoryDialog


class CategoryAddWindow(QDialog, Ui_addCategoryDialog):
    """
    CategoryAddWindow class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init add category window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        # init combo box
        self.parentComboBox.addItem(self._("None"), userData=None)
        for category in self.application.categories.list_all():
            self.parentComboBox.addItem(category.name, userData=category.id)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.nameValueLineEdit.textChanged.connect(self.on_name_line_edit_changed)
        self.buttonBox.rejected.connect(self.close)

    def on_name_line_edit_changed(self) -> None:
        """
        Triggered when the name line edit is changed

        :return:
        """
        name = self.nameValueLineEdit.text().strip()

        if name != "":
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        else:
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        name = self.nameValueLineEdit.text().strip()
        if name == "":
            return None

        parent_id = self.parentComboBox.currentData()

        category = self.application.categories.create(name, parent_id)
        self.application.categories.add(category)

        return None


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    CategoryAddWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import sys

from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QApplication, QMenu, QMessageBox

from tikka.domains.application import Application
from tikka.domains.entities.category import Category
from tikka.domains.entities.constants import DATA_PATH

# if TYPE_CHECKING:
#     from tikka.slots.pyqt.widgets.account_tree import AccountTreeWidget


class CategoryPopupMenu(QMenu):
    """
    CategoryPopupMenu class
    """

    def __init__(
        self,
        application: Application,
        category: Category,
        mutex: QMutex,
        parent: AccountTreeWidget,
    ):
        """
        Init AccountPopupMenu instance

        :param application: Application instance
        :param category: Category instance
        :param mutex: QMutex instance
        :param parent: AccountTreeWidget instance
        """
        super().__init__(parent=parent)

        self._parent = parent
        self.application = application
        self.category = category
        self.mutex = mutex
        self._ = self.application.translator.gettext

        # menu actions
        rename_action = self.addAction(self._("Rename"))
        rename_action.triggered.connect(self.rename)
        add_sub_category_action = self.addAction(self._("Add sub-category"))
        add_sub_category_action.triggered.connect(self.add_sub_category)
        delete_category_action = self.addAction(self._("Delete category"))
        delete_category_action.triggered.connect(self.delete_category)

    def rename(self):
        """
        Set the current category in edit mode in tree view

        :return:
        """
        self._parent.rename_category()

    def add_sub_category(self):
        """
        Add new category in self.category

        :return:
        """
        self._parent.add_sub_category()

    def delete_category(self):
        """
        Add new category in self.category

        :return:
        """
        # display confirm dialog and get response
        button = QMessageBox.question(
            self,
            self._("Delete category"),
            self._(
                'Category content will be deleted locally. Do you really want to delete "{name}"?'
            ).format(name=self.category.name),
        )
        if button == QMessageBox.Yes:
            self.application.categories.delete(self.category.id)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    from tikka.slots.pyqt.widgets.account_tree import AccountTreeWidget

    categories = application_.categories.list(None)
    if len(categories) == 0:
        category_ = application_.categories.create("test category", None)
        application_.categories.add(category_)

    menu = CategoryPopupMenu(
        application_,
        application_.categories.list(None)[0],
        QMutex(),
        AccountTreeWidget(application_, QMutex()),
    )
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Optional

from PyQt5.QtCore import QLocale, QMutex, pyqtSignal
from PyQt5.QtWidgets import QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.technical_committee import TechnicalCommitteeProposal
from tikka.slots.pyqt.entities.constants import (
    TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.widgets.technical_committee_proposal_rc import (
    Ui_TechnicalCommitteeProposalWidget,
)
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class TechnicalCommitteeProposalWidget(QWidget, Ui_TechnicalCommitteeProposalWidget):
    """
    TechnicalCommitteeProposalWidget class
    """

    vote_for_proposal_done = pyqtSignal()

    def __init__(
        self,
        application: Application,
        proposal: TechnicalCommitteeProposal,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init TechnicalCommitteeWidget instance

        :param application: TechnicalCommitteeProposal instance
        :param proposal: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        self.proposal = proposal
        self.vote = False
        self.proposalGroupBox.setTitle(self.proposal.hash)

        self.callIndexValueLabel.setText(str(self.proposal.call.index))
        self.callHashValueLabel.setText(self.proposal.call.hash)
        self.callModuleValueLabel.setText(str(self.proposal.call.module))
        self.callFunctionValueLabel.setText(str(self.proposal.call.function))
        self.callArgumentsValueLabel.setText(str(self.proposal.call.args))

        self.votingIndexValueLabel.setText(str(self.proposal.voting.index))
        self.votingThresholdValueLabel.setText(str(self.proposal.voting.threshold))
        yes_address_names = []
        for address in self.proposal.voting.ayes:
            account = self.application.accounts.get_by_address(address)
            if account is not None:
                yes_address_names.append(
                    f"{address}{' - ' + account.name if account.name is not None else ''}"
                )
        self.votingAyesValueLabel.setText("\n".join(yes_address_names))
        no_address_names = []
        for address in self.proposal.voting.nays:
            account = self.application.accounts.get_by_address(address)
            if account is not None:
                no_address_names.append(
                    f"{address}{' - ' + account.name if account.name is not None else ''}"
                )
        self.votingNaysValueLabel.setText("\n".join(no_address_names))
        end_localized_datetime_string = self.locale().toString(
            self.proposal.voting.end,
            QLocale.dateTimeFormat(self.locale(), QLocale.ShortFormat),
        )
        self.votingEndValueLabel.setText(end_localized_datetime_string)

        # events
        self.yesButton.clicked.connect(self._on_yes_button_clicked)
        self.noButton.clicked.connect(self._on_no_button_clicked)

        # async method
        self.network_vote_async_worker = AsyncQWorker(
            self.network_vote_for_proposal, self.mutex
        )
        self.network_vote_async_worker.finished.connect(
            self._on_network_vote_async_worker_finished
        )

    def _on_yes_button_clicked(self, _):
        """
        Triggered when user click on Yes Button

        :param _:
        :return:
        """
        self.vote = True
        self.network_vote_async_worker.start()

    def _on_no_button_clicked(self, _):
        """
        Triggered when user click on No Button

        :param _:
        :return:
        """
        self.vote = False
        self.network_vote_async_worker.start()

    def network_vote_for_proposal(self):
        """
        Vote and refresh data of technical committee from network

        :return:
        """
        self.errorLabel.setText("")
        preference_account_address_selected = self.application.preferences.get(
            TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_account_address_selected is None:
            self.errorLabel.setText(self._("No technical committee account selected"))
            return

        account = self.application.accounts.get_by_address(
            preference_account_address_selected
        )
        if account is None:
            self.errorLabel.setText(self._("Unknown account"))
            return

        # if account is locked...
        if not self.application.wallets.is_unlocked(
            preference_account_address_selected
        ):
            # ask password...
            dialog_code = AccountUnlockWindow(self.application, account, self).exec_()
            if dialog_code == QDialog.Rejected:
                return

        try:
            self.application.technical_committee.network.vote(
                self.application.wallets.get_keypair(account.address),
                self.proposal,
                self.vote,
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_network_vote_async_worker_finished(self):
        """
        Triggered when async worker is finished

        :return:
        """
        self.vote_for_proposal_done.emit()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QMenu, QMessageBox, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import NodesEvent
from tikka.domains.entities.node import Node


class NodePopupMenu(QMenu):
    """
    NodePopupMenu class
    """

    def __init__(
        self,
        application: Application,
        node: Node,
        parent: Optional[QWidget] = None,
    ):
        """
        Init NodePopupMenu instance

        :param application: Application instance
        :param node: Node instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)

        self.application = application
        self.node = node
        self._ = self.application.translator.gettext

        # menu actions
        copy_url_to_clipboard_action = self.addAction(self._("Copy URL to clipboard"))
        copy_url_to_clipboard_action.triggered.connect(self.copy_url_to_clipboard)
        copy_peer_id_to_clipboard_action = self.addAction(
            self._("Copy peer ID to clipboard")
        )
        copy_peer_id_to_clipboard_action.triggered.connect(
            self.copy_peer_id_to_clipboard
        )
        if (
            not self.node.url == self.application.nodes.get_current_url()
            and self.node.url
            not in self.application.currencies.get_entry_point_urls()["rpc"]
        ):
            forget_node_action = self.addAction(self._("Forget server"))
            forget_node_action.triggered.connect(self.delete_node)

    def copy_peer_id_to_clipboard(self):
        """
        Copy peer ID to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.node.peer_id)

    def copy_url_to_clipboard(self):
        """
        Copy URL to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.node.url)

    def delete_node(self):
        """
        Delete selected node and its entry point

        :return:
        """
        response_button = self.confirm_delete_node(self.node)
        if response_button == QMessageBox.Yes:
            self.application.nodes.delete(self.node.url)
            self.application.event_dispatcher.dispatch_event(
                NodesEvent(NodesEvent.EVENT_TYPE_LIST_CHANGED)
            )

    def confirm_delete_node(self, node: Node) -> QMessageBox.StandardButton:
        """
        Display confirm dialog and return response

        :param node: Node instance
        :return:
        """
        # display confirm dialog and get response
        custom_question = self._("Forget server {}?")
        return QMessageBox.question(
            self,
            self._("Forget server"),
            custom_question.format(node.url),
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    node_ = Node(
        "ws://node.url.com",
        "732SSfuwjB7jkt9th1zerGhphs6nknaCBCTozxUcPWPU",
        999,
        "duniter",
        "1.8.1",
    )

    menu = NodePopupMenu(application_, node_)
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from collections import OrderedDict
from typing import Optional

from mnemonic import Mnemonic
from mnemonic.mnemonic import ConfigurationError
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QButtonGroup, QDialog, QLineEdit, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import AccountCryptoType
from tikka.domains.entities.constants import DATA_PATH, WALLETS_PASSWORD_LENGTH
from tikka.libs.keypair import Keypair
from tikka.libs.secret import generate_alphabetic, sanitize_mnemonic_string
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
)
from tikka.slots.pyqt.resources.gui.windows.account_import_rc import (
    Ui_AccountImportDialog,
)


class AccountImportWindow(QDialog, Ui_AccountImportDialog):
    """
    AccountImportWindow class
    """

    language_code_map = {
        "english": "en",
        "french": "fr",
        "chinese_simplified": "zh-hans",
        "chinese_traditional": "zh-hant",
        "italian": "it",
        "japanese": "ja",
        "korean": "ko",
        "spanish": "es",
    }

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init import account window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.language_code = "en"

        # Mnemonic language selector translated
        self.mnemonic_language_displayed = OrderedDict(
            [
                ("en", self._("English")),
                ("fr", self._("French")),
                ("zh-hans", self._("Chinese simplified")),
                ("zh-hant", self._("Chinese traditional")),
                ("it", self._("Italian")),
                ("ja", self._("Japanese")),
                ("ko", self._("Korean")),
                ("es", self._("Spanish")),
            ]
        )

        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        self.cryptoTypeGroup = QButtonGroup(self)
        self.cryptoTypeGroup.addButton(self.ed25519RadioButton)
        self.cryptoTypeGroup.addButton(self.sr25519RadioButton)
        self.ed25519RadioButton.setChecked(True)
        self.crypto_type = (
            AccountCryptoType.ED25519
            if self.ed25519RadioButton.isChecked()
            else AccountCryptoType.SR25519
        )

        # events
        self.mnemonicLineEdit.textChanged.connect(self._on_mnemonic_changed)
        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.derivationLineEdit.textChanged.connect(self._on_derivation_changed)
        self.cryptoTypeGroup.buttonClicked.connect(self.on_crypto_type_changed)
        self.passwordChangeButton.clicked.connect(self._generate_wallet_password)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        # debounce timers
        self.mnemonic_debounce_timer = QTimer()
        self.mnemonic_debounce_timer.timeout.connect(self._mnemonic_debounce_call)
        self.derivation_debounce_timer = QTimer()
        self.derivation_debounce_timer.timeout.connect(self._derivation_debounce_call)

        # fill form
        self.storedPasswordFrame.hide()
        self._generate_wallet_password()

    def _on_mnemonic_changed(self):
        """
        Triggered when mnemonic is changed

        :return:
        """
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()
        self.mnemonic_debounce_timer.start(DEBOUNCE_TIME)

    def _mnemonic_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing mnemonic

        :return:
        """
        # stop mnemonic_debounce_timer to avoid infinite loop
        if self.mnemonic_debounce_timer.isActive():
            self.mnemonic_debounce_timer.stop()

        self.errorLabel.setText("")

        if self.verify_user_entry() is not True:
            return

        try:
            address = self._generate_address()
        except Exception:
            self.errorLabel.setText(self._("Mnemonic or language not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return

        if self.address_already_exists(address):
            return

        self._update_wallet_password_fields()

        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.mnemonicLineEdit.echoMode() == QLineEdit.Password:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Normal)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.mnemonicLineEdit.setEchoMode(QLineEdit.Password)
            self.storedpasswordLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def _on_derivation_changed(self):
        """
        Triggered when user enter text in derivationLineEdit

        :return:
        """
        if self.derivation_debounce_timer.isActive():
            self.derivation_debounce_timer.stop()
        self.derivation_debounce_timer.start(DEBOUNCE_TIME)

    def _derivation_debounce_call(self):
        """
        Debounce function triggered only after idle time when typing derivation

        :return:
        """
        # stop derivation_debounce_timer to avoid infinite loop
        if self.derivation_debounce_timer.isActive():
            self.derivation_debounce_timer.stop()

        self.errorLabel.setText("")

        try:
            address = self._generate_address()
        except Exception:
            self.errorLabel.setText(self._("Derivation is not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return

        if self.address_already_exists(address):
            return

        self._update_wallet_password_fields()

        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

    def _update_wallet_password_fields(self):
        """
        Show stored password or new password depending on root account exists or not

        :return:
        """
        # if root account exists, hide the password field
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        root_keypair = Keypair.create_from_mnemonic(
            mnemonic=mnemonic,
            language_code=self.language_code,
            crypto_type=self.crypto_type,
            ss58_format=self.application.currencies.get_current().ss58_format,
        )
        if self.application.passwords.exists(root_keypair.ss58_address):
            stored_password = self.application.passwords.get_clear_password(
                root_keypair
            )
            self.storedpasswordLineEdit.setText(stored_password)
            self.storedPasswordFrame.show()
            self.passwordFrame.hide()
        else:
            self.storedPasswordFrame.hide()
            self.passwordFrame.show()

    def on_crypto_type_changed(self):
        """
        Triggered when user click on key type radio button

        :return:
        """
        self.errorLabel.setText("")
        self.crypto_type = (
            AccountCryptoType.ED25519
            if self.ed25519RadioButton.isChecked()
            else AccountCryptoType.SR25519
        )
        if self.mnemonicLineEdit.text().strip() != "":
            address = self._generate_address()
            if not self.address_already_exists(address):
                self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

    def _generate_address(self) -> str:
        """
        Generate, display and return address

        :return:
        """
        # sanitize mnemonic string
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())

        derivation = self.derivationLineEdit.text().strip()
        address = Keypair.create_from_uri(
            suri=mnemonic + derivation,
            ss58_format=self.application.currencies.get_current().ss58_format,
            crypto_type=self.crypto_type,
            language_code=self.language_code,
        ).ss58_address

        self.addressValueLabel.setText(address)
        return address

    def address_already_exists(self, address: str) -> bool:
        """
        Modify form if address already exists

        :param address: Address to check
        :return:
        """
        if self.application.accounts.get_by_address(address) is not None:
            self.errorLabel.setText(self._("Account already exists!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return True

        return False

    def verify_user_entry(self) -> bool:
        """
        Verify user entry, return True if OK, False otherwise

        :return:
        """
        # verify mnemonic
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        try:
            self.language_code = self.language_code_map[
                Mnemonic.detect_language(mnemonic)
            ]
        except ConfigurationError as exception:
            logging.exception(exception)
            self.addressValueLabel.setText("")
            self.errorLabel.setText(self._("Language not detected"))
            self.mnemonicLanguageValueLabel.setText("")
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        if not Keypair.validate_mnemonic(mnemonic, self.language_code):
            self.addressValueLabel.setText("")
            self.errorLabel.setText(self._("Mnemonic not valid!"))
            self.mnemonicLanguageValueLabel.setText("")
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return False

        self.mnemonicLanguageValueLabel.setText(
            self.mnemonic_language_displayed[self.language_code]
        )

        return True

    def _generate_wallet_password(self):
        """
        Generate new password for wallet encryption in UI

        :return:
        """
        self.passwordLineEdit.setText(generate_alphabetic(WALLETS_PASSWORD_LENGTH))

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        # user inputs
        mnemonic = sanitize_mnemonic_string(self.mnemonicLineEdit.text())
        derivation_ = self.derivationLineEdit.text().strip()
        name = self.nameLineEdit.text().strip()

        # generated inputs
        password = self.passwordLineEdit.text()

        self.application.accounts.create_new_account(
            mnemonic, self.language_code, derivation_, self.crypto_type, name, password
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    AccountImportWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Dict, List, Optional

from tikka.domains.entities.identity import Certification, Identity
from tikka.interfaces.adapters.network.node.node import NetworkNodeInterface
from tikka.libs.keypair import Keypair


class NodeIdentitiesInterface(abc.ABC):
    """
    NodeIdentitiesInterface class
    """

    def __init__(self, node: NetworkNodeInterface) -> None:
        """
        Use node connection to request account informations

        :param node: NetworkNodeInterface instance
        :return:
        """
        self.node = node

    @abc.abstractmethod
    def get_identity_index(self, address: str) -> Optional[int]:
        """
        Return the account Identity instance from address if exists

        :param address: Account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identity_indice(self, addresses: List[str]) -> List[Optional[int]]:
        """
        Return the account Identity index from addresses if exists

        :param addresses: Account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identity(self, address: str) -> Optional[Identity]:
        """
        Return the account Identity instance from address if exists

        :param address: Account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identity_by_index(self, index: int) -> Optional[Identity]:
        """
        Return the account Identity instance from index if exists

        :param index: Identity index
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identities(self, addresses: List[str]) -> Dict[str, Optional[Identity]]:
        """
        Return a dict with address: Identity instance from network

        :param addresses: List of account address
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_identities_by_index(
        self, identity_indice: List[int]
    ) -> Dict[int, Optional[Identity]]:
        """
        Return a dict with identity_index: Identity instance from network

        :param identity_indice: List of identity indice
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def change_owner_key(self, old_keypair: Keypair, new_keypair: Keypair) -> None:
        """
        Change identity owner from old_keypair to new_keypair

        :param old_keypair: Keypair of current identity account
        :param new_keypair: Keypair of new identity account
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def certs_by_receiver(
        self, receiver_address: str, receiver_identity_index: int
    ) -> List[Certification]:
        """
        Return a list of certification received by identity_index

         [
         [identity index, expire on block number],
         [identity index, expire on block number]
         ]

        :param receiver_address: Address of account receiving certs
        :param receiver_identity_index: Index of identity receiving certs
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def claim_uds(self, keypair: Keypair) -> None:
        """
        Add unclaimed UDs of identity to keypair account balance

        :param keypair: Keypair of account
        :return:
        """
        raise NotImplementedError


class NodeIdentitiesException(Exception):
    """
    NodeIdentitiesException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QLineEdit, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    DEBOUNCE_TIME,
)
from tikka.slots.pyqt.resources.gui.windows.account_unlock_rc import Ui_UnlockDialog


class AccountUnlockWindow(QDialog, Ui_UnlockDialog):
    """
    AccountUnlockWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        parent: Optional[QWidget] = None,
    ):
        """
        Init unlock account window

        :param application: Application instance
        :param account: Account instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.account = account
        self._ = self.application.translator.gettext

        # populate fields
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)
        self.addressValueLabel.setText(self.account.address)

        # disable Ok button
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.showButton.clicked.connect(self.on_show_button_clicked)
        self.passwordLineEdit.textChanged.connect(
            self._on_password_line_edit_text_changed
        )
        self.buttonBox.rejected.connect(self.on_rejected_button)

        # debounce timer
        self.debounce_timer = QTimer()
        self.debounce_timer.timeout.connect(self._unlock_account)

    def _on_password_line_edit_text_changed(self):
        """
        Triggered when text is changed in the password field

        :return:
        """
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()
        self.debounce_timer.start(DEBOUNCE_TIME)

    def on_show_button_clicked(self):
        """
        Triggered when user click on show button

        :return:
        """
        if self.passwordLineEdit.echoMode() == QLineEdit.Password:
            self.passwordLineEdit.setEchoMode(QLineEdit.Normal)
            self.showButton.setText(self._("Hide"))
        else:
            self.passwordLineEdit.setEchoMode(QLineEdit.Password)
            self.showButton.setText(self._("Show"))

    def _unlock_account(self) -> None:
        """
        Validate fields and unlock account to enabled ok button

        :return:
        """
        # stop debounce_timer to avoid infinite loop
        if self.debounce_timer.isActive():
            self.debounce_timer.stop()

        password = self.passwordLineEdit.text().strip().upper()
        self.errorLabel.setText("")
        try:
            result = self.application.accounts.unlock(self.account, password)
        except Exception:
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Password is not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return None

        if not result:
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Password is not valid!"))
            self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)
            return None

        self.errorLabel.setStyleSheet("color: green;")
        self.errorLabel.setText(self._("Account is unlocked"))
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)
        self.buttonBox.button(self.buttonBox.Ok).click()
        return None

    def on_rejected_button(self) -> None:
        """
        Triggered when user click on cancel button

        :return:
        """
        self.application.accounts.lock(self.account)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account("CUYYUnh7N49WZhs5DULkmqw5Zu5fwsRBmE5LLrUFRpgw")
    AccountUnlockWindow(application_, account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from datetime import datetime
from typing import List, Optional

from tikka.domains.entities.transfer import Transfer
from tikka.interfaces.adapters.network.indexer.indexer import NetworkIndexerInterface


class IndexerTransfersInterface(abc.ABC):
    """
    IndexerTransfersInterface class
    """

    COLUMN_TIMESTAMP = "timestamp"

    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    def __init__(self, indexer: NetworkIndexerInterface) -> None:
        """
        Init IndexerTransfersInterface instance

        :param indexer: NetworkIndexerInterface instance
        :return:
        """
        self.indexer = indexer

    @abc.abstractmethod
    def list(
        self,
        address: str,
        limit: int,
        offset: int = 0,
        sort_column: str = COLUMN_TIMESTAMP,
        sort_order: str = SORT_ORDER_DESCENDING,
        from_datetime: Optional[datetime] = None,
        to_datetime: Optional[datetime] = None,
    ) -> List[Transfer]:
        """
        Return list of transfers from and to address

        :param address: Account address
        :param limit: Max number of transfers to return
        :param offset: Offset to paginate results
        :param sort_column: Sort column, default to IndexerTransfersInterface.COLUMN_TIMESTAMP
        :param sort_order: Sort order, default to IndexerTransfersInterface.SORT_ORDER_DESCENDING
        :param from_datetime: Only transfers from datetime, default to None
        :param to_datetime: Only transfers until datetime, default to None
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def count(self, address) -> int:
        """
        Return transfers total count from and to address

        :param address: Account address
        :return:
        """
        raise NotImplementedError


class IndexerTransfersException(Exception):
    """
    IndexerTransfersException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, List

from tikka.domains.entities.technical_committee import TechnicalCommitteeMember
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.technical_committee_members import (
    TechnicalCommitteeMembersRepositoryInterface,
)

TABLE_NAME = "technical_committee_members"


class DBTechnicalCommitteeMembersRepository(
    TechnicalCommitteeMembersRepositoryInterface, DBRepositoryInterface
):
    """
    DBTechnicalCommitteeMembersRepository class
    """

    def set_list(self, members: List[TechnicalCommitteeMember]) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TechnicalCommitteeMembersRepositoryInterface.set_list.__doc__
        )
        self.client.clear(TABLE_NAME)
        rows = [
            get_fields_from_technical_committee_member(member) for member in members
        ]
        if len(rows) > 0:
            self.client.insert_many(TABLE_NAME, rows)

    def list(self) -> List[TechnicalCommitteeMember]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            TechnicalCommitteeMembersRepositoryInterface.list.__doc__
        )
        result_set = self.client.select(
            f"SELECT * FROM {TABLE_NAME} ORDER BY address ASC"
        )

        list_ = []
        for row in result_set:
            list_.append(get_technical_committee_member_from_row(row))

        return list_


def get_fields_from_technical_committee_member(
    technical_committee_member: TechnicalCommitteeMember,
) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param technical_committee_member: TechnicalCommitteeMember instance
    :return:
    """
    fields = {}
    for (key, value) in technical_committee_member.__dict__.items():
        if key.startswith("_"):
            continue
        fields[key] = value

    return fields


def get_technical_committee_member_from_row(row: tuple) -> TechnicalCommitteeMember:
    """
    Return an TechnicalCommitteeMember instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        values.append(value)
        count += 1

    return TechnicalCommitteeMember(*values)  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/indexer_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_indexerAddDialog(object):
    def setupUi(self, indexerAddDialog):
        indexerAddDialog.setObjectName("indexerAddDialog")
        indexerAddDialog.resize(731, 126)
        self.verticalLayout = QtWidgets.QVBoxLayout(indexerAddDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.urlLabel = QtWidgets.QLabel(indexerAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.urlLabel.setFont(font)
        self.urlLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.urlLabel.setObjectName("urlLabel")
        self.gridLayout.addWidget(self.urlLabel, 0, 0, 1, 1)
        self.blockValueLabel = QtWidgets.QLabel(indexerAddDialog)
        self.blockValueLabel.setText("")
        self.blockValueLabel.setObjectName("blockValueLabel")
        self.gridLayout.addWidget(self.blockValueLabel, 1, 1, 1, 1)
        self.urlLineEdit = QtWidgets.QLineEdit(indexerAddDialog)
        self.urlLineEdit.setObjectName("urlLineEdit")
        self.gridLayout.addWidget(self.urlLineEdit, 0, 1, 1, 1)
        self.testButton = QtWidgets.QPushButton(indexerAddDialog)
        self.testButton.setObjectName("testButton")
        self.gridLayout.addWidget(self.testButton, 0, 2, 1, 1)
        self.blockLabel = QtWidgets.QLabel(indexerAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.blockLabel.setFont(font)
        self.blockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.blockLabel.setObjectName("blockLabel")
        self.gridLayout.addWidget(self.blockLabel, 1, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(indexerAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(indexerAddDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(indexerAddDialog)
        self.buttonBox.accepted.connect(indexerAddDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(indexerAddDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(indexerAddDialog)
        indexerAddDialog.setTabOrder(self.urlLineEdit, self.testButton)

    def retranslateUi(self, indexerAddDialog):
        #  _translate = QtCore.QCoreApplication.translate
        indexerAddDialog.setWindowTitle(_("Add indexer"))
        self.urlLabel.setText(_("URL"))
        self.testButton.setText(_("Test"))
        self.blockLabel.setText(_("Block"))



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Optional

from tikka.adapters.network.node.substrate_client import ExtrinsicReceipt
from tikka.interfaces.adapters.network.node.node import NetworkNodeInterface
from tikka.libs.keypair import Keypair


class NodeTransfersInterface(abc.ABC):
    """
    NodeTransfersInterface class
    """

    def __init__(self, node: NetworkNodeInterface) -> None:
        """
        Use node connection to request account informations

        :param node: NetworkNodeInterface instance
        :return:
        """
        self.node = node

    @abc.abstractmethod
    def send(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send amount (blockchain unit) from sender_account to recipient_address
        wait for extrinsic finalization and return ExtrinsicReceipt

        :param sender_keypair: Sender Keypair
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain units
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def send_with_comment(
        self, sender_keypair: Keypair, recipient_address: str, amount: int, comment: str
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send amount (blockchain unit) from sender_account to recipient_address
        wait for extrinsic finalization and return ExtrinsicReceipt

        :param sender_keypair: Sender Keypair
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain units
        :param comment: Comment from user
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fees(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[int]:
        """
        Fetch transfer fees and return it

        :param sender_keypair: Sender Keypair
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain units
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def transfer_all(
        self, sender_keypair: Keypair, recipient_address: str, keep_alive: bool = False
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send amount (blockchain unit) from sender_account to recipient_address
        wait for extrinsic finalization and return ExtrinsicReceipt

        :param sender_keypair: Sender Keypair
        :param recipient_address: Recipient address
        :param keep_alive: Optional, default False
        :return:
        """
        raise NotImplementedError


class NodeTransfersException(Exception):
    """
    NodeTransfersException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Dict, Optional

from PyQt5.QtCore import QMutex, Qt
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QCompleter, QDialog, QWidget
from scalecodec.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.libs.keypair import Keypair
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_LOADER,
    SELECTED_UNIT_PREFERENCES_KEY,
    TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY,
    TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.transfer_rc import Ui_TransferDialog
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class TransferWindow(QDialog, Ui_TransferDialog):
    """
    TransferWindow class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        sender: Optional[Account] = None,
        recipient: Optional[Account] = None,
        parent: Optional[QWidget] = None,
    ):
        """
        Init transfer window

        :param application: Application instance
        :param sender: Sender Account instance or None
        :param recipient: Recipient Account instance or None
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application

        preferences_sender_address = self.application.repository.preferences.get(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY
        )
        if sender is None and preferences_sender_address is not None:
            preferences_sender_account = self.application.accounts.get_by_address(
                preferences_sender_address
            )
            if preferences_sender_account is not None:
                sender = preferences_sender_account

        preferences_recipient_address = self.application.repository.preferences.get(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY
        )
        if recipient is None and preferences_recipient_address is not None:
            preferences_recipient_account = self.application.accounts.get_by_address(
                preferences_recipient_address
            )
            if preferences_recipient_account is not None:
                recipient = preferences_recipient_account

        self.sender_account = sender
        self.recipient_account = recipient
        self.mutex = mutex
        self._ = self.application.translator.gettext

        # substrate error message for gettext extraction
        self._("Transfer/payment would kill account")
        self._("Balance too low to send value")

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)

        # init unit from preference or default
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.unit = unit_preference
        else:
            self.unit = AMOUNT_UNIT_KEY

        self.init_units()

        self.amount_value = 0
        self.fees = None
        self.transfer_success: Optional[bool] = None

        if self.sender_account is not None:
            if self.sender_account.name is None:
                self.senderNameOrAddressLineEdit.setText(self.sender_account.address)
                self.senderNameOrAddressLineEdit.setFont(self.monospace_font)
                self.senderNameOrAddressValueLabel.setText("")
            else:
                self.senderNameOrAddressLineEdit.setText(self.sender_account.name)
                self.senderNameOrAddressLineEdit.setFont(QFont())
                self.senderNameOrAddressValueLabel.setText(self.sender_account.address)
                self.senderNameOrAddressValueLabel.setFont(self.monospace_font)

        if self.recipient_account is not None:
            if self.recipient_account.name is None:
                self.recipientNameOrAddressLineEdit.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText("")
            else:
                self.recipientNameOrAddressLineEdit.setText(self.recipient_account.name)
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                self.recipientNameOrAddressValueLabel.setText(
                    self.recipient_account.address
                )
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        # autocomplete  (dirty version)
        # todo: create a model to fetch only a filtered list from database on edit
        self.account_address_by_name: Dict[str, str] = {}
        account_address_by_name_duplicate_count = 1
        for account_instance in self.application.accounts.get_list():
            if (
                account_instance.name in self.account_address_by_name
                and account_instance.name is not None
            ):
                self.account_address_by_name[
                    f"{account_instance.name} ({account_address_by_name_duplicate_count})"
                ] = account_instance.address
            else:
                if account_instance.name is None:
                    self.account_address_by_name[""] = account_instance.address
                else:
                    self.account_address_by_name[
                        account_instance.name
                    ] = account_instance.address

        # sender autocomplete only on accounts with wallet
        self.sender_account_wordlist = []
        for account_name, account_address in self.account_address_by_name.items():
            if account_address in self.application.wallets.list_addresses():
                self.sender_account_wordlist.append(account_name)
                self.sender_account_wordlist.append(account_address)
        sender_completer = QCompleter(self.sender_account_wordlist)
        self.senderNameOrAddressLineEdit.setCompleter(sender_completer)
        self.senderNameOrAddressLineEdit.completer().setCaseSensitivity(
            Qt.CaseInsensitive
        )

        # recipient autocomplete
        self.recipient_account_wordlist = list(
            self.account_address_by_name.keys()
        ) + list(self.account_address_by_name.values())
        recipient_completer = QCompleter(self.recipient_account_wordlist)
        self.recipientNameOrAddressLineEdit.setCompleter(recipient_completer)
        self.recipientNameOrAddressLineEdit.completer().setCaseSensitivity(
            Qt.CaseInsensitive
        )

        self._update_ui()

        # events
        self.senderNameOrAddressLineEdit.textChanged.connect(
            self._on_sender_address_line_edit_changed
        )
        self.recipientNameOrAddressLineEdit.textChanged.connect(
            self._on_recipient_address_line_edit_changed
        )
        self.amountDoubleSpinBox.valueChanged.connect(
            self._on_amount_double_spin_box_changed
        )
        self.amountUnitComboBox.activated.connect(self._on_unit_changed)
        self.feesButton.clicked.connect(self._on_fees_button_clicked)
        self.sendButton.clicked.connect(self._on_send_button_clicked)
        self.buttonBox.button(self.buttonBox.Close).clicked.connect(self.close)

        ##############################
        # ASYNC METHODS
        ##############################
        # fetch recipient balance
        self.fetch_recipient_balance_from_network_async_qworker = AsyncQWorker(
            self.fetch_recipient_balance_from_network, self.mutex, self
        )
        self.fetch_recipient_balance_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_recipient_balance_from_network
        )
        # fetch sender balance
        self.fetch_sender_balance_from_network_async_qworker = AsyncQWorker(
            self.fetch_sender_balance_from_network, self.mutex, self
        )
        self.fetch_sender_balance_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_sender_balance_from_network
        )

        # fetch fees from network
        self.fetch_fees_from_network_async_qworker = AsyncQWorker(
            self.fetch_fees_from_network, self.mutex, self
        )
        self.fetch_fees_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_fees_from_network
        )
        # send transfer to network
        self.send_tranfer_to_network_async_qworker = AsyncQWorker(
            self.send_transfer_to_network, self.mutex, self
        )
        self.send_tranfer_to_network_async_qworker.finished.connect(
            self._on_finished_send_transfer_to_network
        )

        # Receiver is a history account, not stored...
        if (
            self.recipient_account is not None
            and self.application.accounts.get_by_address(self.recipient_account.address)
            is None
        ):
            # get balance from node
            self.fetch_recipient_balance_from_network_async_qworker.start()

    def _on_unit_changed(self):
        """
        Triggered when unit combo box is changed

        :return:
        """
        self.unit = self.amountUnitComboBox.currentData()
        self._update_ui()

    def _on_amount_double_spin_box_changed(self):
        """
        Triggered when the amount spin box is changed

        :return:
        """
        self.amount_value = self.amountDoubleSpinBox.value()
        self._update_ui()

    def fetch_recipient_balance_from_network(self):
        """
        Fetch last account data from the network

        :return:
        """
        self.loaderIconLabel.show()
        try:
            # if account not in stored accounts...
            if (
                self.application.accounts.get_by_address(self.recipient_account.address)
                is None
            ):
                self.recipient_account.balance = (
                    self.application.accounts.network_get_balance(
                        self.recipient_account.address
                    )
                )
            else:
                self.recipient_account = (
                    self.application.accounts.network_update_balance(
                        self.recipient_account
                    )
                )
        except Exception as exception:
            # keep balance from cache DB if network failure
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_recipient_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def fetch_sender_balance_from_network(self):
        """
        Fetch last sender account data from the network

        :return:
        """
        self.loaderIconLabel.show()

        try:
            self.sender_account = self.application.accounts.network_update_balance(
                self.sender_account
            )
        except Exception as exception:
            # keep balance from cache DB if network failure
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_sender_balance_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self._update_ui()

    def _on_sender_address_line_edit_changed(self):
        """
        Triggered when text in the sender address field is changed

        :return:
        """
        self.errorLabel.setText("")
        address_or_name = self.senderNameOrAddressLineEdit.text().strip()

        try:
            address_is_valid = is_valid_ss58_address(address_or_name)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            # if entry is a known account name...
            if address_or_name in self.account_address_by_name:
                self.senderNameOrAddressLineEdit.setFont(QFont())
                # display address under the name
                self.senderNameOrAddressValueLabel.setFont(self.monospace_font)
                self.senderNameOrAddressValueLabel.setText(
                    self.account_address_by_name[address_or_name]
                )
                self._get_sender_account(self.account_address_by_name[address_or_name])
            else:
                self.senderNameOrAddressValueLabel.setText("")
                self.senderBalanceValueLabel.setText("")
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Invalid address! Please check it again.")
                )
                self.sender_account = None

        else:
            self.senderNameOrAddressLineEdit.setFont(self.monospace_font)
            # display name under the address
            for key, value in self.account_address_by_name.items():
                if value == address_or_name:
                    self.senderNameOrAddressValueLabel.setFont(QFont())
                    self.senderNameOrAddressValueLabel.setText(key)
            self._get_sender_account(address_or_name)

        self._update_ui()

    def _on_recipient_address_line_edit_changed(self):
        """
        Triggered when text in the recipient address field is changed

        :return:
        """
        self.errorLabel.setText("")
        address_or_name = self.recipientNameOrAddressLineEdit.text().strip()

        try:
            address_is_valid = is_valid_ss58_address(address_or_name)
        except IndexError:
            address_is_valid = False

        if not address_is_valid:
            # if entry is a known account name...
            if address_or_name in self.account_address_by_name:
                self.recipientNameOrAddressLineEdit.setFont(QFont())
                # display address under the name
                self.recipientNameOrAddressValueLabel.setFont(self.monospace_font)
                self.recipientNameOrAddressValueLabel.setText(
                    self.account_address_by_name[address_or_name]
                )
                self._get_recipient_account(
                    self.account_address_by_name[address_or_name]
                )
            else:
                self.recipientNameOrAddressValueLabel.setText("")
                self.recipientBalanceValueLabel.setText("")
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Invalid address! Please check it again.")
                )
                self.recipient_account = None
        else:
            self.recipientNameOrAddressLineEdit.setFont(self.monospace_font)
            # display name under the address
            for key, value in self.account_address_by_name.items():
                if value == address_or_name:
                    self.recipientNameOrAddressValueLabel.setFont(QFont())
                    self.recipientNameOrAddressValueLabel.setText(key)
            self._get_recipient_account(address_or_name)

        self._update_ui()

    def _get_sender_account(self, address: str) -> None:
        """
        Get sender account

        :return:
        """
        self.sender_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.transfer_success = False
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.sender_account = None
            return None

        # save in preferences
        self.application.repository.preferences.set(
            TRANSFER_SENDER_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.sender_account = account
                self._update_ui()
                return None

        # create account
        self.sender_account = Account(address)
        # fetch balance
        self.fetch_sender_balance_from_network_async_qworker.start()

        return None

    def _get_recipient_account(self, address: str) -> None:
        """
        Get recipient account

        :return:
        """
        self.recipient_account = None
        try:
            Keypair(
                ss58_address=address,
                ss58_format=self.application.currencies.get_current().ss58_format,
            )
        except ValueError as exception:
            logging.exception(exception)
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid address! Please check it again."))
            self.recipient_account = None
            return None

        # save in preferences
        self.application.repository.preferences.set(
            TRANSFER_RECIPIENT_ADDRESS_PREFERENCES_KEY, address
        )

        # search in local account list
        for account in self.application.accounts.get_list():
            if account.address == address:
                self.recipient_account = account
                self._update_ui()
                return None

        # create account
        self.recipient_account = Account(address)
        # fetch balance
        self.fetch_recipient_balance_from_network_async_qworker.start()

        return None

    def _update_ui(self):
        """
        Update UI

        :return:
        """
        # balances amount
        amount = self.application.amounts.get_amount(self.unit)
        if self.sender_account is not None:
            if self.sender_account.balance is None:
                self.senderBalanceValueLabel.setText("?")
            else:
                self.senderBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.sender_account.balance), amount.symbol()
                    )
                )

        if self.recipient_account is not None:
            if self.recipient_account.balance is None:
                self.recipientBalanceValueLabel.setText("?")
                self.recipientBalanceValueLabel.setToolTip(
                    self._(
                        "Account balance unknown! Send only one unit and make sure the owner can get it"
                    )
                )
            else:
                self.recipientBalanceValueLabel.setText(
                    self.locale().toCurrencyString(
                        amount.value(self.recipient_account.balance), amount.symbol()
                    )
                )
                self.recipientBalanceValueLabel.setToolTip("")

        # fees amount
        if self.fees is not None:
            self.feesValueLabel.setText(
                self.locale().toCurrencyString(amount.value(self.fees), amount.symbol())
            )
        else:
            self.feesValueLabel.setText("?")

        # buttons
        if (
            self.sender_account is not None
            and self.recipient_account is not None
            and self.amount_value > 0
        ):
            self.sendButton.setDisabled(False)
            self.feesButton.setDisabled(False)
        else:
            self.feesButton.setDisabled(True)
            self.sendButton.setDisabled(True)

    def init_units(self) -> None:
        """
        Init units combobox for transfer amount

        :return:
        """
        self.amountUnitComboBox.clear()

        for key, amount in self.application.amounts.register.items():
            self.amountUnitComboBox.addItem(amount.name(), userData=key)
        self.amountUnitComboBox.setCurrentIndex(
            self.amountUnitComboBox.findData(self.unit)
        )

    def _on_fees_button_clicked(self):
        """
        Triggered when user click on Fees button

        :return:
        """
        # sender account locked...
        if not self.application.wallets.is_unlocked(self.sender_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.sender_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.feesButton.setDisabled(True)
        self.loaderIconLabel.show()

        self.fetch_fees_from_network_async_qworker.start()

    def fetch_fees_from_network(self):
        """
        Fetch fees amount from network

        :return:
        """
        amount = self.application.amounts.get_amount(self.unit)
        # get value as blockchain units
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        try:
            self.fees = self.application.transfers.network_fees(
                sender_keypair, self.recipient_account.address, blockchain_value
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_fees_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.feesButton.setDisabled(False)
        self._update_ui()

    def _on_send_button_clicked(self):
        """
        Triggered when user click on Send button

        :return:
        """
        # sender account locked...
        if not self.application.wallets.is_unlocked(self.sender_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.sender_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.sendButton.setDisabled(True)
        self.loaderIconLabel.show()
        # send transfer to network
        self.send_tranfer_to_network_async_qworker.start()

    def send_transfer_to_network(self):
        """
        Send transfer to network

        :return:
        """
        self.errorLabel.setText("")
        # get value as blockchain units
        amount = self.application.amounts.get_amount(self.unit)
        blockchain_value = amount.blockchain_value(self.amountDoubleSpinBox.value())
        sender_keypair = self.application.wallets.get_keypair(
            self.sender_account.address
        )
        comment = self.commentLineEdit.text().strip()

        try:
            ascii_comment = comment.encode("ascii")
        except UnicodeEncodeError:
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("Invalid character in comment"))
            return

        if sender_keypair is None:
            return

        try:
            if comment == "":
                extrinsic_receipt = self.application.transfers.network_send(
                    sender_keypair, self.recipient_account.address, blockchain_value
                )
            else:
                extrinsic_receipt = (
                    self.application.transfers.network_send_with_comment(
                        sender_keypair,
                        self.recipient_account.address,
                        blockchain_value,
                        ascii_comment,
                    )
                )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            if extrinsic_receipt is None:
                self.transfer_success = False
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(
                    self._("Transfer failed. Please check logs to understand why")
                )
            elif extrinsic_receipt.is_success is False:
                self.transfer_success = False
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(self._(extrinsic_receipt.error_message["name"]))
            else:
                self.transfer_success = True
                self.errorLabel.setStyleSheet("color: green;")
                self.errorLabel.setText(self._("Transfer done"))

    def _on_finished_send_transfer_to_network(self):
        """
        Triggered when async request send_transfer_to_network is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.sendButton.setDisabled(False)
        self.fetch_sender_balance_from_network_async_qworker.start()
        self.fetch_recipient_balance_from_network_async_qworker.start()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)

    account1 = application_.accounts.create_new_root_account(
        "album cute glance oppose hub fury strategy health dust rebuild trophy magic",
        "en",
        AccountCryptoType.ED25519,
        "test account",
        "aaaaaa",
    )
    if account1 is not None:
        account1.balance = 1000
    account2 = application_.accounts.create_new_account(
        "album cute glance oppose hub fury strategy health dust rebuild trophy magic",
        "en",
        "//1",
        AccountCryptoType.ED25519,
        "test account",
        "aaaaaa",
    )
    if account2 is not None:
        account2.balance = 0

    TransferWindow(application_, QMutex(), account1, account2).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.config import Config
from tikka.domains.entities.constants import DATA_PATH, LOCALES_PATH
from tikka.slots.pyqt.resources.gui.widgets.licence_rc import Ui_LicenceWidget


class LicenceWidget(QWidget, Ui_LicenceWidget):
    """
    LicenceWidget class
    """

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init licence widget

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        with open(
            LOCALES_PATH.joinpath(
                application.config.get(Config.LANGUAGE_KEY), "licence_g1.md"
            ),
            "r",
            encoding="utf-8",
        ) as input_file:
            markdown_text = input_file.read()
        self.textBrowser.setMarkdown(markdown_text)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(LicenceWidget(application_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pathlib import Path
from typing import Optional

from PyQt5.QtCore import QDate
from PyQt5.QtWidgets import QDialog, QFileDialog, QWidget

from tikka.domains.application import Application
from tikka.slots.pyqt.entities.constants import (
    TRANSFERS_EXPORT_DEFAULT_DIRECTORY_PREFERENCES_KEY,
)
from tikka.slots.pyqt.resources.gui.windows.transfers_export_rc import (
    Ui_transfersExportDialog,
)


class TransfersExportWindow(QDialog, Ui_transfersExportDialog):
    """
    TransfersExportWindow class
    """

    MIN_PERIOD_IN_WEEKS = 4
    MAX_PERIOD_IN_WEEKS = 104

    def __init__(
        self, application: Application, address: str, parent: Optional[QWidget] = None
    ):
        """
        Init transfers export window instance

        :param application: Application instance
        :param address: Account address
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.address = address

        # buttons
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)
        self.fromCalendarWidget.selectionChanged.connect(self.update_date_constraints)
        self.toCalendarWidget.selectionChanged.connect(self.update_date_constraints)
        self.todayButton.clicked.connect(self.init_calendar_limits)
        self.init_calendar_limits()

    def init_calendar_limits(self):
        """
        Définit les dates minimales et maximales autorisées pour les calendriers,
        en se basant sur la période maximale définie.
        """
        today = QDate.currentDate()
        default_min_date = today.addDays(-self.MIN_PERIOD_IN_WEEKS * 7)

        # Définir la plage de dates pour le calendrier "from"
        # self.fromCalendarWidget.setDateRange(max_min_date, today)
        self.fromCalendarWidget.setSelectedDate(default_min_date)

        # Définir la plage de dates pour le calendrier "to"
        # self.toCalendarWidget.setDateRange(min_date, max_date)
        self.toCalendarWidget.setSelectedDate(today)

    def update_date_constraints(self):
        """
        Update calendars with date constraints
        """
        from_date = self.fromCalendarWidget.selectedDate()
        to_date = self.toCalendarWidget.selectedDate()
        min_to_date = from_date.addDays(self.MIN_PERIOD_IN_WEEKS * 7)
        max_from_date = to_date.addDays(-self.MAX_PERIOD_IN_WEEKS * 7)

        if from_date > to_date:
            self.fromCalendarWidget.setSelectedDate(to_date)
        elif to_date < min_to_date:
            self.toCalendarWidget.setSelectedDate(min_to_date)

        max_to_date = from_date.addDays(self.MAX_PERIOD_IN_WEEKS * 7)

        if to_date > max_to_date:
            self.toCalendarWidget.setSelectedDate(max_to_date)
        if from_date < max_from_date:
            self.fromCalendarWidget.setSelectedDate(max_from_date)
        # self.toCalendarWidget.setMinimumDate(from_date)
        # self.toCalendarWidget.setMaximumDate(max_to_date)
        # self.fromCalendarWidget.setMaximumDate(to_date)

    def open_file_dialog(self) -> Optional[str]:
        """
        Open file dialog and return the selected filepath or None

        :return:
        """
        default_dir = self.application.repository.preferences.get(
            TRANSFERS_EXPORT_DEFAULT_DIRECTORY_PREFERENCES_KEY
        )
        if default_dir is not None:
            default_dir = str(Path(default_dir).expanduser().absolute())
        else:
            default_dir = ""

        result = QFileDialog.getSaveFileName(
            self, self._("Export file"), default_dir, "OFX Files (*.ofx)"
        )
        if result[0] == "":
            return None

        self.application.repository.preferences.set(
            TRANSFERS_EXPORT_DEFAULT_DIRECTORY_PREFERENCES_KEY,
            str(Path(result[0]).parent),
        )

        return result[0]

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        # open file dialog
        filepath = self.open_file_dialog()
        if filepath is not None:
            self.application.transfers.export_as_ofx(
                filepath,
                self.address,
                self.fromCalendarWidget.selectedDate().toPyDate(),
                self.toCalendarWidget.selectedDate().toPyDate(),
            )
        self.close()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import json
import logging
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any
from uuid import UUID

from tikka.adapters.network.network import Network
from tikka.adapters.repository.repository import Repository
from tikka.domains.accounts import Accounts
from tikka.domains.amounts import Amounts
from tikka.domains.authorities import Authorities
from tikka.domains.categories import Categories
from tikka.domains.config import Config
from tikka.domains.connections import Connections
from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account
from tikka.domains.entities.category import Category
from tikka.domains.entities.constants import DATABASE_FILE_EXTENSION, LOCALES_PATH
from tikka.domains.entities.events import CurrencyEvent
from tikka.domains.entities.indexer import Indexer
from tikka.domains.entities.node import Node
from tikka.domains.entities.password import Password
from tikka.domains.entities.wallet import Wallet
from tikka.domains.events import EventDispatcher
from tikka.domains.identities import Identities
from tikka.domains.indexers import Indexers
from tikka.domains.nodes import Nodes
from tikka.domains.passwords import Passwords
from tikka.domains.preferences import Preferences
from tikka.domains.smiths import Smiths
from tikka.domains.technical_committee import TechnicalCommittee
from tikka.domains.transfers import Transfers
from tikka.domains.vaults import Vaults
from tikka.domains.wallets import Wallets
from tikka.interfaces.adapters.network.node.currency import NodeCurrencyException


class Application:
    """
    Application class
    """

    def __init__(self, data_path: Path):
        """
        Init application

        :param data_path: Path instance of application data folder
        """

        # set data path
        self.data_path = data_path

        # dependency injection
        # init event dispatcher
        self.event_dispatcher = EventDispatcher()

        # repository adapter
        self.repository = Repository(str(self.data_path))

        # init config domain
        self.config = Config(self.repository.config)

        # if supported currencies has changed and config use an unknown currency...
        if (
            self.config.get(Config.CURRENCY_KEY)
            not in self.repository.currencies.code_names()
        ):
            # set current currency to first currency by default
            self.config.set(
                Config.CURRENCY_KEY, self.repository.currencies.code_names()[0]
            )

        # database adapter
        currency_db_file_uri = (
            self.data_path.expanduser()
            .joinpath(self.config.get(Config.CURRENCY_KEY) + DATABASE_FILE_EXTENSION)
            .as_uri()
        )
        self.repository.db_client.connect(currency_db_file_uri)

        # network adapter
        self.network = Network()

        # init translation
        self.translator = self.init_i18n()

        # init domains
        self.connections = Connections(self.network, self.event_dispatcher)
        self.preferences = Preferences(self.repository.preferences)
        self.currencies = Currencies(
            self.repository.currencies,
            self.repository.currency,
            self.network.node.currency,
            self.config.get(Config.CURRENCY_KEY),
        )
        self.passwords = Passwords(self.repository.passwords)
        self.wallets = Wallets(self.repository.wallets, self.currencies)
        self.transfers = Transfers(
            self.wallets,
            self.repository.transfers,
            self.currencies,
            self.network.node.transfers,
            self.network.indexer.transfers,
            self.event_dispatcher,
        )
        self.identities = Identities(
            self.repository.identities,
            self.network.node.identities,
            self.network.indexer.identities,
        )
        self.accounts = Accounts(
            self.repository.accounts,
            self.network.node.accounts,
            self.network.indexer.accounts,
            self.passwords,
            self.wallets,
            self.transfers,
            self.repository.v1_file_wallets,
            self.currencies,
            self.event_dispatcher,
        )
        self.amounts = Amounts(self.currencies, self.translator)
        self.nodes = Nodes(
            self.repository.nodes,
            self.preferences,
            self.network.node,
            self.config,
            self.currencies,
            self.event_dispatcher,
        )
        self.indexers = Indexers(
            self.repository.indexers,
            self.preferences,
            self.network.indexer,
            self.config,
            self.currencies,
            self.event_dispatcher,
        )
        self.smiths = Smiths(
            self.repository.smiths,
            self.repository.identities,
            self.network.node,
            self.network.indexer.identities,
        )
        self.authorities = Authorities(
            self.repository.authorities,
            self.network.node.authorities,
            self.nodes,
            self.smiths,
        )
        self.categories = Categories(
            self.repository.categories, self.accounts, self.event_dispatcher
        )
        self.technical_committee = TechnicalCommittee(
            self.repository,
            self.network,
            self.event_dispatcher,
        )
        self.vaults = Vaults(
            self.network.node.accounts,
            self.accounts,
            self.currencies,
        )

        # if currency properties required for amount display not populated...
        if self.currencies.get_current().members_count is None:
            try:
                # fetch currency properties from network
                self.currencies.network_update_properties()
            except NodeCurrencyException:
                self.currencies.get_current().members_count = 1

    def init_i18n(self) -> Any:
        """
        Init translator from configured language

        :return:
        """
        # define translator for configurated language
        translator = gettext.translation(
            "application",
            str(LOCALES_PATH),
            languages=[self.config.get(Config.LANGUAGE_KEY)],
        )
        # init translator
        translator.install()

        return translator

    def select_currency(self, code_name: str):
        """
        Change currency

        :return:
        """
        if self.config is None:
            raise NoConfigError

        # dispatch event EVENT_TYPE_PRE_CHANGE
        event = CurrencyEvent(CurrencyEvent.EVENT_TYPE_PRE_CHANGE, code_name)
        self.event_dispatcher.dispatch_event(event)

        self.config.set(Config.CURRENCY_KEY, code_name)

        if self.repository.db_client is not None:
            self.repository.db_client.disconnect()

        # init database connection
        currency_db_file_uri = (
            self.data_path.expanduser()
            .joinpath(code_name + DATABASE_FILE_EXTENSION)
            .as_uri()
        )
        self.repository.db_client.connect(currency_db_file_uri)

        self.currencies.set_current(code_name)

        # init domains with new repository adapter
        self.nodes.init_repository()
        self.indexers.init_repository()

        # get current entry point for new network connection
        current_node = self.nodes.get(self.nodes.get_current_url())
        if current_node is not None:
            # disconnect previous node connection
            self.connections.node.disconnect()
            # connect to node
            self.connections.node.connect(current_node)

        # get current entry point for new network connection
        current_indexer = self.indexers.get(self.indexers.get_current_url())
        if current_indexer is not None:
            # disconnect previous indexer connection
            self.connections.indexer.disconnect()
            # connect to indexer
            self.connections.indexer.connect(current_indexer)

        try:
            # fetch currency properties from network
            self.currencies.network_update_properties()
        except NodeCurrencyException:
            self.currencies.get_current().members_count = 1

        # dispatch event EVENT_TYPE_CHANGED
        event = CurrencyEvent(
            CurrencyEvent.EVENT_TYPE_CHANGED, self.currencies.get_current().code_name
        )
        self.event_dispatcher.dispatch_event(event)

    def select_language(self, language: str):
        """
        Select GUI language

        :param language: Code of language (ex: "en_US", "fr_FR")
        :return:
        """
        if self.config is None:
            raise NoConfigError

        self.config.set(Config.LANGUAGE_KEY, language)
        self.translator = self.init_i18n()

    def save_data(self, filepath: str) -> None:
        """
        Save user data on disk as json file

        :param filepath: Filepath to save data into
        :return:
        """
        categories = []
        for category in self.categories.list_all():
            category_dict = asdict(category)
            category_dict = serialize_dict(category_dict)
            categories.append(category_dict)

        accounts = []
        for account in self.accounts.get_list():
            account_dict = asdict(account)
            account_dict = serialize_dict(account_dict)
            accounts.append(account_dict)

        wallets = []
        for wallet in self.wallets.list():
            wallets.append(asdict(wallet))

        passwords = []
        for password in self.passwords.list():
            passwords.append(asdict(password))

        nodes = []
        for node in self.nodes.list():
            nodes.append(asdict(node))

        indexers = []
        for indexer in self.indexers.list():
            indexers.append(asdict(indexer))

        data = {
            "categories": categories,
            "accounts": accounts,
            "wallets": wallets,
            "passwords": passwords,
            "nodes": nodes,
            "indexers": indexers,
        }
        with open(filepath, "w") as file:
            json.dump(data, file, indent=4)

        logging.debug("Repository data successfully saved in %s", filepath)

    def load_data(self, filepath: str) -> None:
        """
        Load backup data json from disk

        :param filepath: Path of json file
        :return:
        """
        logging.debug("Delete all data in repository...")
        self.categories.delete_all()
        self.accounts.delete_all()
        self.wallets.delete_all()
        self.passwords.delete_all()
        self.identities.delete_all()
        self.smiths.delete_all()
        self.authorities.delete_all()

        self.nodes.delete_all()
        self.nodes.init_repository()
        self.indexers.delete_all()
        self.indexers.init_repository()

        logging.debug("Open json file...")
        with open(filepath, "r") as file:
            data = json.load(file)

        logging.debug("process user data...")
        for category_dict in data["categories"]:
            category_dict["id"] = UUID(hex=category_dict["id"])
            self.categories.repository.add(Category(**category_dict))

        for account_dict in data["accounts"]:
            account_dict["category_id"] = (
                UUID(hex=account_dict["category_id"])
                if account_dict["category_id"] is not None
                else None
            )
            self.accounts.repository.add(Account(**account_dict))

        for wallet_dict in data["wallets"]:
            self.wallets.repository.add(Wallet(**wallet_dict))

        for password_dict in data["passwords"]:
            self.passwords.repository.add(Password(**password_dict))

        for node_dict in data["nodes"]:
            self.nodes.repository.add(Node(**node_dict))

        for indexer_dict in data["indexers"]:
            self.indexers.repository.add(Indexer(**indexer_dict))

        logging.debug("User data successfully loaded")

        # load currency servers in repository
        self.nodes.init_repository()
        self.indexers.init_repository()

    def close(self):
        """
        Quit application and close what needs to be closed

        :return:
        """
        # disconnect all connections
        self.connections.disconnect_all()
        # close DB client
        self.repository.db_client.disconnect()


def serialize_dict(data: dict) -> dict:
    """
    Return the data dict with serialized values for json export

    :param data: Data to serialize
    :return:
    """
    serialized_data = {}
    for key, value in data.items():
        if isinstance(value, UUID):
            value = value.hex
        elif isinstance(value, datetime):
            value = value.isoformat()
        serialized_data[key] = value

    return serialized_data


class NoDatabaseError(Exception):
    pass


class NoConfigError(Exception):
    pass



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from gettext import NullTranslations

from tikka.domains.currencies import Currencies
from tikka.domains.entities.currency import Currency


class AmountInterface(abc.ABC):

    currency: Currency

    def __init__(self, currencies: Currencies, translator: NullTranslations):
        """
        Init AmountInterface class with blockchain value (normally centimes)

        :param currencies: Currencies domain instance
        :param translator: NullTranslations instance (gettext)
        """
        self.currencies = currencies
        self.translator = translator

        self._ = self.translator.gettext

    @abc.abstractmethod
    def value(self, blockchain_value: int) -> float:
        """
        Return amount value expressed in class unit

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def blockchain_value(self, value: float) -> int:
        """
        Return blockchain value from value in class unit

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def symbol(self) -> str:
        """
        Return amount unit symbol

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def name(self) -> str:
        """
        Return amount unit name

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, List, Optional

from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.identities import (
    IdentitiesRepositoryInterface,
)

TABLE_NAME = "identities"


class DBIdentitiesRepository(IdentitiesRepositoryInterface, DBRepositoryInterface):
    """
    DBIdentitiesRepository class
    """

    def add(self, identity: Identity) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_identity(identity),
        )

    def get(self, index: int) -> Optional[Identity]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.get.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE index_=?", (index,)
        )
        if row is None:
            return None

        return get_identity_from_row(row)

    def get_by_address(self, address: str) -> Optional[Identity]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.get_by_address.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE address=?", (address,)
        )
        if row is None:
            return None

        return get_identity_from_row(row)

    def get_index_by_address(self, address: str) -> Optional[int]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.get_index_by_address.__doc__
        )

        row = self.client.select_one(
            f"SELECT index_ FROM {TABLE_NAME} WHERE address=?", (address,)
        )
        if row is None:
            return None

        return row[0]

    def update(self, identity: Identity) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"index_='{identity.index}'",
            **get_fields_from_identity(identity),
        )

    def delete(self, index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, index_=index)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.delete_all.__doc__
        )

        self.client.clear(TABLE_NAME)

    def exists(self, index: int) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.exists.__doc__
        )

        row = self.client.select_one(
            f"SELECT count(index_) FROM {TABLE_NAME} WHERE index_=?", (index,)
        )
        if row is None:
            return False

        return row[0] == 1

    def list(self) -> List[Identity]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.list.__doc__
        )

        result_set = self.client.select(f"SELECT * FROM {TABLE_NAME}")
        list_ = []
        for row in result_set:
            list_.append(get_identity_from_row(row))

        return list_

    def list_indice(self) -> List[int]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IdentitiesRepositoryInterface.list_indice.__doc__
        )

        result_set = self.client.select(f"SELECT index_ FROM {TABLE_NAME}")
        list_indice = []
        for row in result_set:
            list_indice.append(row[0])

        return list_indice


def get_fields_from_identity(identity: Identity) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param identity: Identity instance
    :return:
    """
    fields = {}
    for (key, value) in identity.__dict__.items():
        if key.startswith("_"):
            continue
        if key == "index":
            # index is a reserved keyword in sqlite3
            key = "index_"
        if key == "status":
            # convert IdentityStatus Enum to int
            value = value.value
        fields[key] = value

    return fields


def get_identity_from_row(row: tuple) -> Identity:
    """
    Return an Identity instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 3:
            # convert int to IdentityStatus Enum
            values.append(IdentityStatus(value))
        else:
            values.append(value)
        count += 1

    return Identity(*values)  # pylint: disable=no-value-for-parameter



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QMenu, QMessageBox, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import IndexersEvent
from tikka.domains.entities.indexer import Indexer


class IndexerPopupMenu(QMenu):
    """
    IndexerPopupMenu class
    """

    def __init__(
        self,
        application: Application,
        indexer: Indexer,
        parent: Optional[QWidget] = None,
    ):
        """
        Init IndexerPopupMenu instance

        :param application: Application instance
        :param indexer: Indexer instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)

        self.application = application
        self.indexer = indexer
        self._ = self.application.translator.gettext

        # menu actions
        copy_url_to_clipboard_action = self.addAction(self._("Copy URL to clipboard"))
        copy_url_to_clipboard_action.triggered.connect(self.copy_url_to_clipboard)
        if (
            not self.indexer.url == self.application.indexers.get_current_url()
            and self.indexer.url
            not in self.application.currencies.get_entry_point_urls()["squid"]
        ):
            forget_indexer_action = self.addAction(self._("Forget server"))
            forget_indexer_action.triggered.connect(self.delete_indexer)

    def copy_url_to_clipboard(self):
        """
        Copy URL to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.indexer.url)

    def delete_indexer(self):
        """
        Forget selected indexer and its entry point

        :return:
        """
        response_button = self.confirm_delete_indexer(self.indexer)
        if response_button == QMessageBox.Yes:
            self.application.indexers.delete(self.indexer.url)
            self.application.event_dispatcher.dispatch_event(
                IndexersEvent(IndexersEvent.EVENT_TYPE_LIST_CHANGED)
            )

    def confirm_delete_indexer(self, indexer: Indexer) -> QMessageBox.StandardButton:
        """
        Display confirm dialog and return response

        :param indexer: Indexer instance
        :return:
        """
        # display confirm dialog and get response
        custom_question = self._("Forget server {}?")
        return QMessageBox.question(
            self,
            self._("Forget server"),
            custom_question.format(indexer.url),
        )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    indexer_ = Indexer(
        "ws://indexer.url.com",
        999,
    )

    menu = IndexerPopupMenu(application_, indexer_)
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional

from tikka.adapters.network.node.substrate_client import ExtrinsicReceipt
from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account
from tikka.domains.entities.events import TransferEvent
from tikka.domains.entities.transfer import Transfer
from tikka.domains.events import EventDispatcher
from tikka.domains.wallets import Wallets
from tikka.interfaces.adapters.network.indexer.transfers import (
    IndexerTransfersInterface,
)
from tikka.interfaces.adapters.network.node.transfers import NodeTransfersInterface
from tikka.interfaces.adapters.repository.transfers import TransfersRepositoryInterface
from tikka.libs.keypair import Keypair


def format_datetime(dt: datetime) -> str:
    """
    Formate une date en OFX (YYYYMMDDHHMMSS).

    :param dt: Datetime data
    :return:
    """
    return dt.strftime("%Y%m%d%H%M%S")


class Transfers:
    """
    Transfers domain class
    """

    def __init__(
        self,
        wallets: Wallets,
        repository: TransfersRepositoryInterface,
        currencies: Currencies,
        node_transfers: NodeTransfersInterface,
        indexer_transfers: IndexerTransfersInterface,
        event_dispatcher: EventDispatcher,
    ):
        """
        Init Transfers domain

        :param wallets: Wallets domain
        :param repository: TransfersRepositoryInterface adapter instance
        :param currencies: Currencies domain instance
        :param node_transfers: NodeTransfersInterface adapter instance
        :param event_dispatcher: EventDispatcher instance
        """
        self.wallets = wallets
        self.repository = repository
        self.currencies = currencies
        self.node_transfers = node_transfers
        self.indexer_transfers = indexer_transfers
        self.event_dispatcher = event_dispatcher

    def add(self, address: str, transfer: Transfer):
        """
        Add transfer to repository

        :param transfer: Transfer instance
        :return:
        """
        self.repository.add(address, transfer)

    def delete(self, transfer_id: str):
        """
        Delete transfer with transfer_id in repository

        :param transfer_id: Transfer ID to delete
        :return:
        """
        self.repository.delete(transfer_id)

    def delete_all(self) -> None:
        """
        Delete all transfers in repository

        :return:
        """
        self.repository.delete_all()

    def count(self, address: str) -> int:
        """
        Return total number of transfers issued and received by address

        :return:
        """
        return self.repository.count(address)

    def list(
        self,
        address: str,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: str = TransfersRepositoryInterface.COLUMN_TIMESTAMP,
        sort_order: str = TransfersRepositoryInterface.SORT_ORDER_DESCENDING,
        limit: int = 10,
        offset: int = 0,
    ) -> List[Transfer]:
        """
        List transfers from and to address, from repository with optional filters and sort_column

        :param address: Account address
        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant, TransfersRepositoryInterface.COLUMN_TIMESTAMP by default
        :param sort_order: Sort order constant, TransfersRepositoryInterface.SORT_ORDER_DESCENDING by default
        :param limit: Number of rows to return
        :param offset: Offset of first row to return
        :return:
        """
        return self.repository.list(
            address, filters, sort_column, sort_order, limit, offset
        )

    def network_fetch_history_for_account(self, account: Account, limit: int) -> None:
        """
        Fetch history of last transfers from and to account

        :param account: Account instance to filter issuers and receivers
        :param limit: Maximum number of transfers to fetch
        :return:
        """
        transfers = self.indexer_transfers.list(account.address, limit=limit)
        if len(transfers) > 0:
            self.repository.set_history(account.address, transfers)
            # account.last_transfer_timestamp = transfers[0].timestamp
            # # dispatch event
            # event = LastTransferEvent(
            #     LastTransferEvent.EVENT_TYPE_LAST_TRANSFER_CHANGED,
            #     account.address,
            #     transfers[0],
            # )
            # self.event_dispatcher.dispatch_event(event)

    def network_fetch_total_count_for_address(self, address: str) -> int:
        """
        Fetch total count of transfers from and to address

        :param address: Address to filter issuers and receivers
        :return:
        """
        return self.indexer_transfers.count(address)

    def network_fees(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[int]:
        """
        Fetch transfer fees from network and return it if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain unit
        :return:
        """
        return self.node_transfers.fees(sender_keypair, recipient_address, amount)

    def network_send(
        self, sender_keypair: Keypair, recipient_address: str, amount: int
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send transfer to network and return ExtrinsicReceipt if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain unit
        :return:
        """
        receipt = self.node_transfers.send(sender_keypair, recipient_address, amount)

        if receipt is not None and receipt.is_success is True:
            # dispatch event
            event = TransferEvent(
                TransferEvent.EVENT_TYPE_SENT,
            )
            self.event_dispatcher.dispatch_event(event)

        return receipt

    def network_send_with_comment(
        self, sender_keypair: Keypair, recipient_address: str, amount: int, comment: str
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send transfer to network and return ExtrinsicReceipt if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param amount: Amount in blockchain unit
        :param comment: Comment from user
        :return:
        """
        receipt = self.node_transfers.send_with_comment(
            sender_keypair, recipient_address, amount, comment
        )

        if receipt is not None and receipt.is_success is True:
            # dispatch event
            event = TransferEvent(
                TransferEvent.EVENT_TYPE_SENT,
            )
            self.event_dispatcher.dispatch_event(event)

        return receipt

    def network_send_all(
        self, sender_keypair: Keypair, recipient_address: str, keep_alive: bool = False
    ) -> Optional[ExtrinsicReceipt]:
        """
        Send transfer to network and return ExtrinsicReceipt if request is successful

        :param sender_keypair: Sender Keypair instance
        :param recipient_address: Recipient address
        :param keep_alive: Optional, default False
        :return:
        """
        receipt = self.node_transfers.transfer_all(
            sender_keypair, recipient_address, keep_alive
        )

        if receipt is not None and receipt.is_success is True:
            # dispatch event
            event = TransferEvent(
                TransferEvent.EVENT_TYPE_SENT,
            )
            self.event_dispatcher.dispatch_event(event)

        return receipt

    def export_as_ofx(
        self,
        filepath: str,
        address: str,
        from_datetime: Optional[datetime] = None,
        to_datetime: Optional[datetime] = None,
    ):
        """
        Export transfers for address in an OFX format file

        :param filepath: Path of the file
        :param address: Account address
        :param from_datetime: Optional period start date
        :param to_datetime: Optional period end date
        :return:
        """
        batch_size = 100
        offset = 0

        with open(filepath, "w", encoding="utf-8") as f:
            # Écriture de l'en-tête OFX
            f.write(
                f"""OFXHEADER:100
    DATA:OFXSGML
    VERSION:102
    SECURITY:NONE
    ENCODING:USASCII
    CHARSET:1252
    COMPRESSION:NONE
    OLDFILEUID:NONE
    NEWFILEUID:NONE
    
    <OFX>
        <BANKMSGSRSV1>
            <STMTTRNRS>
                <STMTRS>
                    <CURDEF>EUR</CURDEF>
                    <BANKACCTFROM>
                        <BANKID>{self.currencies.get_current().name}</BANKID>
                        <ACCTID>{address}</ACCTID>
                        <ACCTTYPE>CHECKING</ACCTTYPE>
                    </BANKACCTFROM>
                    <BANKTRANLIST>
    """
            )

            while True:
                # Récupérer un lot de 100 transferts depuis GraphQL
                transfers = self.indexer_transfers.list(
                    address=address,
                    limit=batch_size,
                    offset=offset,
                    from_datetime=from_datetime,
                    to_datetime=to_datetime,
                )

                if not transfers:
                    break  # Arrêter si plus de transferts à récupérer

                total_amount: float = 0.0
                for transfer in transfers:
                    if transfer.issuer_address != address:
                        amount_sign = 1
                        name = (
                            f"{transfer.issuer_address} - {transfer.issuer_identity_name}#{transfer.issuer_identity_index}"
                            if transfer.issuer_identity_name
                            else transfer.issuer_address
                        )
                    else:
                        amount_sign = -1
                        name = (
                            f"{transfer.receiver_address} - {transfer.receiver_identity_name}#{transfer.receiver_identity_index}"
                            if transfer.receiver_identity_name
                            else transfer.receiver_address
                        )
                    total_amount += (
                        transfer.amount / 100
                    ) * amount_sign  # Convert cents to dollars

                    f.write(
                        f"""
                    <STMTTRN>
                        <TRNTYPE>XFER</TRNTYPE>
                        <DTPOSTED>{format_datetime(transfer.timestamp)}</DTPOSTED>
                        <TRNAMT>{(transfer.amount / 100)*amount_sign:.2f}</TRNAMT>
                        <FITID>{transfer.id}</FITID>
                        <NAME>{name}</NAME>
                        <MEMO>{transfer.comment or ''}</MEMO>
                    </STMTTRN>
                    """
                    )

                offset += batch_size  # Passer au lot suivant

            current_datetime = format_datetime(datetime.utcnow())

            # Écriture du pied de fichier OFX
            f.write(
                f"""
                    </BANKTRANLIST>
                    <LEDGERBAL>
                        <BALAMT>{total_amount:.2f}</BALAMT>
                        <DTASOF>{current_datetime}</DTASOF>
                    </LEDGERBAL>
                    <AVAILBAL>
                        <BALAMT>{total_amount:.2f}</BALAMT>
                        <DTASOF>{current_datetime}</DTASOF>
                    </AVAILBAL>
                </STMTRS>
            </STMTTRNRS>
        </BANKMSGSRSV1>
    </OFX>
    """
            )

        logging.debug(f"Export OFX terminé : {filepath}")
